require 'gosu'

def load_data(filename)
  File.open(filename, 'rb') { |file| return Marshal.load(file) }
end

def save_data(obj, filename)
  File.open(filename, 'wb') { |file| Marshal.dump(obj, file) }
end

class RGSSError < StandardError
end

class Module
  private
  def cattr_reader(*syms)
    singleton = class << self; self end
    syms.each do |sym|
      singleton.send :define_method, sym, lambda {
        class_variable_get("@@#{sym}")
      }
    end
  end
  def cattr_writer(*syms)
    singleton = class << self; self end
    syms.each do |sym|
      singleton.send :define_method, "#{sym}=", lambda { |val|
        class_variable_set("@@#{sym}", val)
      }
    end
  end
  def cattr_accessor(*syms)
    cattr_reader(*syms)
    cattr_writer(*syms)
  end
end

# Manager class
# Instances of this class keep track of a series of objects (which may
# optionally be sorted by block; see Enumerable#sort!) and may delegate certain
# actions to those objects.
class Manager
  def initialize(&sort)
    @managed = []
    @sort = sort
  end
  def manage(*objects)
    @managed |= objects
  end
  def unmanage(*objects)
    @managed.clear if objects.first == :all
    @managed -= objects
  end
  def delegate(manage_method = nil, &block)
    @managed.sort!(&@sort)
    if block_given?
      @managed.each(&block)
      return
    end
    return if manage_method.nil?
    @managed.each do |object|
      object.__send__(manage_method) if object.respond_to?(manage_method)
    end
  end
end

module Settings
  SettingsFile = './game.ini'
  
  SEARCH_PATHS = ['./']
  
  @data = Hash.new { |hash, key| hash[key] = [] }
  
  IO.readlines(SettingsFile).each do |line|
    if (ar = line.chomp.split('=', 2)).size == 2
      @data[ar[0]] << ar[1] 
    end
  end

  # Get cached value; latest value only
  def self.[](key)
    return '' unless @data.has_key?(key)
    @data[key].last
  end
end

class GosuWindow < Gosu::Window
  attr_reader   :draw_manager
  attr_reader   :showing_fps
  def initialize(*args)
    @draw_manager = Manager.new { |a, b| a.z <=> b.z }
    @drawn = false
    @showing_fps = false
    super(*args)
    recaption
  end
  def showing_fps=(bool)
    @showing_fps = bool
    recaption
    bool
  end
  def recaption
    to_show = Settings['Title']
    to_show += ' (%d fps)' % [Gosu.fps] if @showing_fps
    self.caption = to_show
  end
  def update
    Thread.stop if @drawn
  end
  def draw(draw_graphics = true)
    @drawn = true
    c = 0xff000000
    draw_quad(0,0,c, width,0,c, 0,height,c, width,height,c)
    @draw_manager.delegate(:draw)
    Graphics.draw if draw_graphics
  end
  def needs_redraw?
    !Graphics.skipping?
  end
end

module Disposable
  def dispose
    @disposed = true
  end
  def disposed?
    @disposed == true
  end
end

class Rect
  attr_accessor :x, :y, :width, :height
  def initialize(x, y, width, height)
    @x, @y, @width, @height = x, y, width, height
  end
  def set(x, y, width, height)
    @x, @y, @width, @height = x, y, width, height
  end
  def empty
    @x = @y = @width = @height = 0
  end
  def to_a(coords = false)
    return [@x, @y, @x + @width - 1, @y + @height - 1] if coords
    [@x, @y, @width, @height]
  end
  def ==(other)
    return false unless other.is_a? Rect
    return false unless @x == other.x and @y == other.y
    @width == other.width and @height == other.height
  end
end
