class Tone
  attr_reader :red, :green, :blue, :gray
  def initialize(red, green, blue, gray = 0)
    self.red, self.green, self.blue, self.gray = red, green, blue, gray
  end
  def red=(value)
    @red = [[-255, value].max, 255].min
  end
  def green=(value)
    @green = [[-255, value].max, 255].min
  end
  def blue=(value)
    @blue = [[-255, value].max, 255].min
  end
  def gray=(value)
    @gray = [[0, value].max, 255].min
  end
  def blend(tone)
    self.clone.blend!(tone)
  end
  def blend!(tone)
    self.red += tone.red
    self.green += tone.green
    self.blue += tone.blue
    self.gray += tone.gray
    self
  end
  # Dump to String (compatible with RGSS Tone dump format)
  def _dump(marshal_depth = -1)
    [@red, @green, @blue, @gray].pack('E4')
  end
  # Load from String (compatible with RGSS Tone dump format)
  def self._load(data)
    new(*data.unpack('E4'))
  end
end

#TODO: Waves
class Sprite
  include Comparable # Comparable by :z values
  include Disposable
  DEFAULTS = {:bitmap => nil, :src_rect => nil, :viewport => nil,
              :visible => true, :x => 0, :y => 0, :z => 0, :ox => 0, :oy => 0,
              :zoom_x => 1.0, :zoom_y => 1.0, :angle => 0, :wave_amp => 0,
              :wave_length => 0, :wave_speed => 360, :wave_phase => 0,
              :mirror => false, :bush_depth => 0, :bush_opacity => 128,
              :opacity => 255, :blend_type => 0}
  BLEND_TYPES = [:default, :add, :subtract]
  attr_reader   :bitmap, :viewport, :gosuwindow
  attr_accessor :src_rect, :visible, :x, :y, :z, :ox, :oy, :zoom_x, :zoom_y
  attr_accessor :angle, :wave_amp, :wave_length, :wave_speed, :wave_phase
  attr_accessor :mirror, :bush_depth, :bush_opacity, :opacity, :blend_type
  attr_accessor :color, :tone
  def initialize(viewport = nil, opts = {})
    @gosuwindow = $GameWindow
    @color = Color::Clear.dup
    @tone = Tone.new(0, 0, 0)
    DEFAULTS.merge!(opts).each do |sym, val|
      instance_variable_set("@#{sym}", val)
    end
    (@viewport.nil? ? @gosuwindow.draw_manager : @viewport).manage(self)
  end
  def bitmap=(bitmap)
    @bitmap = bitmap
    @src_rect = bitmap.rect if !@bitmap.nil? and @src_rect.nil?
    @cached_image = bitmap.dup
    bitmap
  end
  def viewport=(viewport)
    (@viewport.nil? ? @gosuwindow.draw_manager : @viewport).unmanage(self)
    (viewport.nil? ? @gosuwindow.draw_manager : viewport).manage(self)
    @viewport = viewport
  end
  def gosuwindow=(gosuwindow)
    if @viewport.nil?
      @gosuwindow.draw_manager.unmanage(self)
      gosuwindow.draw_manager.manage(self)
    end
    @gosuwindow = gosuwindow
  end
  def dispose
    (@viewport.nil? ? @gosuwindow.draw_manager : @viewport).unmanage(self)
    @viewport = nil
    super
  end
  def flash(color, duration)
    if color.nil?
      return unless @visible
      color = Color::Normal
      @visible = false
    end
    @flash_color = color
    @flash_minus = color.alpha / duration.to_f
  end
  def update
    unless @flash_color.nil?
      @flash_color.alpha -= @flash_minus
      @flash_color = nil if @flash_color.alpha <= 0
    end
    @wave_phase += @wave_speed / Graphics.frame_rate if @wave_amp > 0
  end
  def width
    @src_rect.width
  end
  def height
    @src_rect.height
  end
  def recache
    @recache = cache_components
    return if @bitmap.nil? or @src_rect.nil?
    @cached_image.paint {
      trect(0, 0, @cached_image.width - 1, @cached_image.height - 1,
           :fill => true, :color => :alpha)
      splice(@bitmap, @src_rect.x, @src_rect.y, :crop => @src_rect.to_a(true))
      rcolor, rtone = real_color, real_tone
      if rcolor.alpha > 0
        trect(0, 0, @cached_image.width - 1, @cached_image.height - 1,
             :alpha_blend => true, :color_control => lambda do |c|
          c[0], c[1] = rcolor.red / 255.0, rcolor.green / 255.0
          c[2], c[3] = rcolor.blue / 255.0, c[3] * rcolor.alpha / 255.0
          c
        end)
      end
#      @cached_image.each do |c, x, y|
#        color = Color.new(c[0] * 255 + rtone.red, c[1] * 255 + rtone.green,
#                          c[2] * 255 + rtone.blue, c[3] * 255)
#        color.saturation -= rtone.gray / 255.0 if rtone.gray > 0
#        c[0..3] = *color.to_a
#        c[3] *= @bush_opacity / 255.0 if @src_rect.height - y < @bush_depth
#      end
    }
  end
  def cache_components
    return nil if @bitmap.nil?
    [@gosuwindow, @bitmap.to_blob, @src_rect.to_a, #@wave_amp, @wave_length, @wave_speed, @wave_phase,
     @bush_depth, @bush_opacity, real_color, real_tone]
  end
  def real_color
    output = @color.clone
    output.blend!(@flash_color) unless @flash_color.nil?
    output.blend!(@viewport.real_color) unless @viewport.nil?
    output
  end
  def real_tone
    return @tone if @viewport.nil?
    @tone.blend(@viewport.real_tone)
  end
  def draw
    return unless @visible
    recache if @recache != cache_components
    return if @cached_image.nil?
    gcolor = Gosu::Color.argb(@opacity, 255, 255, 255)
    cx = (@ox + @src_rect.x) / @cached_image.width.to_f
    cy = (@oy + @src_rect.y) / @cached_image.height.to_f
    zx = (@mirror ? -@zoom_x : @zoom_x)
    @cached_image.draw_rot(@x, @y, @z, @angle, cx, cy, zx, @zoom_y, gcolor,
                           BLEND_TYPES[@blend_type])
  end
  def <=>(other)
    return 0 unless other.respond_to? :z
    @z <=> other.z
  end
end

# Note that while this inherits a lot of functionality from Sprite, some things
# are ignored: @x, @y, @angle, and @mirror (which could theoretically work but
# since they are not in the original interface it's not worth it).
# Other things, like bush depth, still work but not like you might hope for.
# Wave effects still work but take careful planning to keep smooth at the
# boundary where one copy of the image ends and another begins.
class Plane < Sprite
  def draw
    return unless @visible
    recache if @recache != cache_components
    return if @cached_image.nil?
    # Max width and height
    parent = (@viewport.nil? ? @gosuwindow : @viewport.rect)
    w = parent.width  + @cached_image.width
    h = parent.height + @cached_image.height
    # Starting X and Y
    start_x, start_y = @ox, @oy
    start_x += @cached_image.width while start_x < 0
    start_y += @cached_image.height while start_y < 0
    start_x %= @cache_image.width
    start_y %= @cached_image.height
    start_x, start_y = start_x * -@zoom_x, start_y * -@zoom_y
    @cached_image.draw_tiled(start_x, start_y, w, h, @z, @zoom_x, @zoom_y,
                             gcolor, BLEND_TYPES[@blend_type])
  end
end

# Viewport class
# Controls clipping, scrolling, zooming, and rotating of many objects.
class Viewport < Manager
  include Comparable # Comparable by :z
  include Disposable
  DEFAULTS = {:viewport => nil, :visible => true, :z => 0, :ox => 0, :oy => 0,
              :zoom_x => 1.0, :zoom_y => 1.0, :center_x => 0, :center_y => 0}
  attr_accessor :rect, :visible, :z, :ox, :oy, :color, :tone
  attr_accessor :zoom_x, :zoom_y, :center_x, :center_y
  attr_reader   :viewport, :gosuwindow
  def initialize(*args)
    opts = (args.last.is_a?(Hash) ? args.pop : {})
    @gosuwindow = $GameWindow
    @color = Color::Clear.dup
    @tone = Tone.new(0, 0, 0)
    @rect = Graphics.rect
    DEFAULTS.merge!(opts).each do |sym, val|
      instance_variable_set("@#{sym}", val)
    end
    @rect = Rect.new(*args) if args.size == 4
    @rect = args[0] if args[0].is_a?(Rect)
    (@viewport.nil? ? @gosuwindow.draw_manager : @viewport).manage(self)
    super()
  end
  def viewport=(viewport)
    (@viewport.nil? ? @gosuwindow.draw_manager : @viewport).unmanage(self)
    (viewport.nil? ? @gosuwindow.draw_manager : viewport).manage(self)
    @viewport = viewport
  end
  def gosuwindow=(gosuwindow)
    if @viewport.nil?
      @gosuwindow.draw_manager.unmanage(self)
      gosuwindow.draw_manager.manage(self)
    end
    delegate { |object| object.gosuwindow = gosuwindow }
    @gosuwindow = gosuwindow
  end
  def flash(color, duration)
    if color.nil?
      return unless @visible
      color = Color::Normal
      @visible = false
    end
    @flash_color = color
    @flash_minus = color.alpha / duration.to_f
  end
  def update
    unless @flash_color.nil?
      @flash_color.alpha -= @flash_minus
      @flash_color = nil if @flash_color.alpha <= 0
    end
  end
  def dispose
    unmanage(:all)
    (@viewport.nil? ? @gosuwindow.draw_manager : @viewport).unmanage(self)
    @viewport = nil
    super
  end
  def real_color
    output = @color.clone
    output.blend!(@flash_color) unless @flash_color.nil?
    output.blend!(@viewport.real_color) unless @viewport.nil?
    output
  end
  def real_tone
    return @tone if @viewport.nil?
    @tone.blend(@viewport.real_tone)
  end
  def draw
    return if !@visible
    @gosuwindow.flush
    @gosuwindow.clip_to(*@rect.to_a) do
      @gosuwindow.scale(@zoom_x,@zoom_y,@rect.x+@center_x,@rect.y+@center_y) do
        @gosuwindow.translate(@rect.x - @ox, @rect.y - @oy) do
          delegate(:draw)
        end
      end
    end
    @gosuwindow.flush
  end
  def <=>(other)
    return 0 unless other.respond_to? :z
    @z <=> other.z
  end
end

module Graphics
  cattr_accessor :frame_count, :brightness
  cattr_reader   :frame_rate, :width, :height, :fullscreen, :frozen
  @@width = 640
  @@height = 480
  @@fullscreen = false
  @@frame_count = 0
  @@frame_rate = 40
  @@brightness = 255
  @@frozen = true
  @@frozen_image = nil
  @@frozen_opacity = 0
  @@skipping = false
  def self.update
    Audio.update
    @@window_thread.wakeup
    sleep(0.001) until @@window_thread.stop?
    exit unless @@window_thread.alive?
    @@frame_count += 1
  end
  def self.wait(duration)
    duration.times { update }
  end
  def self.fadeout(duration)
    diff = self.brightness / duration
    duration.times do
      self.brightness -= diff
      self.update
    end
  end
  def self.fadein(duration)
    diff = self.brightness / duration
    duration.times do
      self.brightness += diff
      self.update
    end
  end
  def self.freeze
    @@frozen = true
    # Create the frozen image; don't use snap_to_bitmap since we don't need
    # to draw the window without graphics elements
    @@frozen_image = Bitmap.new(@@width, @@height)
    texinfo = @@frozen_image.gl_tex_info
    x = (texinfo.left * Gosu::MAX_TEXTURE_SIZE).to_i
    y = (texinfo.top * Gosu::MAX_TEXTURE_SIZE).to_i
    $GameWindow.to_texture(texinfo.tex_name, x, y, 0, 0, @@width, @@height)
    @@frozen_opacity = 255.0
  end
  def self.transition(duration = 8, filename = nil, vague = 40)
    freeze unless @@frozen
    @@frozen = false
    if filename.nil?
      diff = @@frozen_opacity / duration
      duration.times do
        @@frozen_opacity -= diff
        self.update
      end
    else
      #TODO: File-based transitions
    end
  end
  def self.snap_to_bitmap
    image = Bitmap.new(@@width, @@height)
    $GameWindow.render_to_image(image) { $GameWindow.draw(false) }
    image
  end
  def self.frame_reset
    @@skipping = true
  end
  def self.skipping?
    output = @@skipping
    @@skipping = false
    output
  end
  def self.resize_screen(width, height)
    @@width, @@height = width, height
    recreate
  end
  def self.frame_rate=(rate)
    @@frame_rate = rate
    recreate
    @@frame_rate
  end
  def self.fullscreen=(fullscreen)
    @@fullscreen = fullscreen
    recreate
    @@fullscreen
  end
  def self.rect
    Rect.new(0, 0, @@width, @@height)
  end
  def self.recreate
    @@active = false
    $GameWindow.close
    @@window_thread.join
    old_window = $GameWindow
    create
    old_window.draw_manager.delegate { |obj| obj.gosuwindow = $GameWindow }
  end
  def self.create
    interval = 1000.0 / @@frame_rate
    $GameWindow = GosuWindow.new(@@width, @@height, @@fullscreen, interval)
    @@window_thread = Thread.new { $GameWindow.show }
    @@window_thread.abort_on_exception = true
  end
  def self.draw
    c = Gosu::Color.argb(255 - @@brightness, 0, 0, 0)
    $GameWindow.draw_quad(0,0,c, @@width,0,c, 0,@@height,c, @@width,@@height,c)
    unless @@frozen_image.nil? or @@frozen_opacity == 0
      c = Gosu::Color.argb(@@frozen_opacity.to_i, 255, 255, 255)
      # Frozen image is upside down from to_texture
      @@frozen_image.draw(0, @@height, 0, 1.0, -1.0, c)
    end
  end
  create
end
